/*
 * Fast simultaneous reductions and Chinese remaindering
 *
 * November 2006, Versailles, France.
 * G. Lecerf
 *
 */

declare verbose fastcrt_verbose, 5;

intrinsic Fanin(x::[RngElt]:THRESHOLD:=1)->SeqEnum
{
 
  x is a sequence of ring elements.
 
 
   It returns a sequence l of sequences of the divide and conquer
   products x[i]:
   let m=#l, n=#x and r=THRESHOLD then
   l[1] = [ x[1]...x[r], ..., x[r*(i-1)+1]...x[r*i],
          x[r*i+1]...x[n]],
   l[i] = [ l[i-1][1]*l[i-1][2], l[i-1][3]*l[i-1][4], .... ],...,
   l[m] = [x[1]...x[n]].
 
}
  n:=#x;
  a:=n div THRESHOLD;
  L:=[ &*[x[i] : i in [THRESHOLD*j+1..(j+1)*THRESHOLD]]: j in [0..a-1]];
  if n gt THRESHOLD*a then
    Append(~L,&*[x[i] : i in [THRESHOLD*a+1..n]]);
    a+:=1;
  end if;
  l:=[L];
  while a gt 1 do
    last:=L[a];
    b := a div 2;
    L:= [ L[2*i+1]*L[2*(i+1)] : i in [0..b-1]];
    if (a mod 2 ) eq 1 then
      Append(~L,last);
      b+:=1;
    end if;
    a:=b;
    Append(~l,L);
  end while;
  return l;
end intrinsic;

intrinsic FaninDown(v::RngElt,fanin)->SeqEnum
{
 
  v is an Euclidien ring element.
  fanin is returned by Fanin.
 
 
   It returns the sequence of values [ v mod z : z in ^fanin[#fanin]];
 
}
  L:=[v];
  for i:=#fanin to 1 by -1 do
    Q:=fanin[i];
    L:=[L[((j-1) div 2)+1] mod Q[j]: j in [1..#Q]];
  end for;
  return(L);
end intrinsic;

intrinsic FaninUp(v::[RngElt],fanin)->RngElt
{
 
  v is an Euclidien ring element.
  fanin is returned by Fanin.
 
 
   It returns sum_i v[i] prod_j<>i fanin[j].
 
}
  P:=v;
  for F in fanin do
    last:=P[#P];
    P:= [ P[2*i+1]*F[2*(i+1)] + F[2*i+1]*P[2*(i+1)] : i in [0..(#F div 2)-1]];
    if (#F mod 2 ) eq 1 then Append(~P,last); end if;
  end for;
  assert #P eq 1;
  return P[1];
end intrinsic;

simultaneous_evaluate:=function(f,x,fanin,THRESHOLD)
  UP:=Universe(f);
  T:=UP.1;
  R:=BaseRing(UP);
  ans:=[];
  for e in f do
    tmp:=FaninDown(e,fanin);
    i:=1;
    val:=[R| ];
    for l:=1 to #fanin[1] do
      for j := i to Min(#x,i+THRESHOLD-1) do
        Append(~val,Evaluate(tmp[l],x[j]));
      end for;
      i+:=THRESHOLD;
    end for;
    Append(~ans,val);
  end for;
  return ans;
end function;

intrinsic SimultaneousEvaluate(f::[RngUPolElt],x::[RngElt]:THRESHOLD:=7)-> SeqEnum
{
}
  ;
  UP:=Universe(f);
  T:=UP.1;
  R:=BaseRing(UP);
  fanin:=Fanin([UP| T - e : e in x]:THRESHOLD:=THRESHOLD);
  ans:=simultaneous_evaluate(f,x,fanin,THRESHOLD);
  ;
  return ans;
end intrinsic;

intrinsic SimultaneousEvaluate(f::RngUPolElt,x::[RngElt]:THRESHOLD:=7)-> SeqEnum
{
 Evaluate f with the given values of x substituted
 for the variable of f.
}
  return SimultaneousEvaluate([f],x:THRESHOLD:=THRESHOLD)[1];
end intrinsic;

intrinsic FastInterpolation(P::[RngElt], V::[RngElt]:THRESHOLD:=7) -> RngUPolElt
{
 Given n points P and n values V from a
 field K, return the unique polynomial f over K
 of degree less than n such that
 f(P[i]) = V[i] for each i.
}
  ;
  F:=Universe(V);
  UP<T>:=PolynomialAlgebra(F);
  n:=#P;
  fanin:=Fanin([UP| T - e : e in P]:THRESHOLD:=THRESHOLD);
  L:=fanin[#fanin];
  w:=simultaneous_evaluate([Derivative(L[1])],P,fanin,THRESHOLD)[1];
  ans:=[];
  i:=0;
  for f in fanin[1] do
    Append(~ans,Interpolation(
         [P[i+j]: j in [1..Degree(f)]],
         [V[i+j]*Evaluate(Derivative(f),P[i+j])/w[i+j]: j in [1..Degree(f)]]));
    i+:=Degree(f);
  end for;
  ans:= FaninUp(ans,fanin);
  ;
  return ans; 
end intrinsic;

intrinsic SimultaneousModulo(p::[RngElt],x::[RngElt]:THRESHOLD:=64)-> SeqEnum
{
}
  ;
  fanin:=Fanin(x:THRESHOLD:=THRESHOLD);
  ans:=[];
  R:=Universe(x);
  for e in p do
    tmp:=FaninDown(e,fanin);
    i:=1;
    val:=[R| ];
    for l:=1 to #fanin[1] do
      for j := i to Min(#x,i+THRESHOLD-1) do
        Append(~val,tmp[l] mod x[j]);
      end for;
      i+:=THRESHOLD;
    end for;
    Append(~ans,val);
  end for;
  ;
  return ans;
end intrinsic;

intrinsic SimultaneousModulo(p::RngElt,x::[RngElt]:THRESHOLD:=32)-> SeqEnum
{
 It returns the sequence [p mod z: z in x].
}
  return SimultaneousModulo([p],x:THRESHOLD:=THRESHOLD)[1];
end intrinsic;

intrinsic FastChineseRemainderTheorem(X::[RngElt],M::[RngElt])->RngElt,RngElt
{
 A solution x, &*M of the system of simultaneous linear congruences
 defined by
 the integer sequences X and M so that x = X[i] mod M[i] for each i
 (error if no solution) [#X = #M, M[i] > 0 for all i].
}
  n:=#X;
  require #M eq n: "arguments must have the same size";
  R:=Universe(X);
  fanin:=Fanin(M:THRESHOLD:=1);
  w:=SimultaneousModulo(fanin[#fanin][1],[ e^2 : e in M]:THRESHOLD:=1);
  w:=[InverseMod(w[i] div M[i],M[i]) : i in [1..n]];
  p:=[X[i]*w[i] mod M[i]: i in [1..#M]];
  return FaninUp(p,fanin) mod fanin[#fanin][1], fanin[#fanin][1];
end intrinsic;

intrinsic FastChineseRemainderTheorem_numerator(X::[RngElt], M::[RngElt])->RngElt,RngElt
{
  Returns N=sum_i X[i] prod_j<>i M[j], &*M.
}
  n:=#M;
  fanin:=Fanin(M:THRESHOLD:=1);
  return FaninUp(X,fanin),fanin[#fanin][1];
end intrinsic;
